package gov.fnal.xpathsearch.cache.xmldb;

import com.sleepycat.dbxml.*;
import com.sleepycat.db.TransactionConfig;

import java.io.Writer;
import java.io.Reader;
import java.io.StringReader;
import java.util.HashMap;

public class AccessXMLDB {

    private static XmlContainerConfig containerCfg;
    private static XmlDocumentConfig docConfig;
    private static TransactionConfig txnConfig;
    private static HashMap<String, XmlContainer> XmlContainerMap = new HashMap();

    public static void update(String docString, String docName, String dbName) throws Exception {
        XmlTransaction txn = XMLDB.XMLDBFactory.getXmlManager().createTransaction();
        boolean completed = false;
        try {
            XmlContainer container = getContainer(dbName);
            XmlDocument doc = container.getDocument(txn, docName);
            doc.setContent(docString); //Will replace content saved
            container.updateDocument(txn, doc, null); //Last arg is context.
            completed = true;
        } catch (XmlException xmle) {
            if (xmle.getErrorCode()==XmlException.DOCUMENT_NOT_FOUND) {
                add(docString, docName, dbName);
            } else {
                System.err.println("Could not update document.");
                xmle.printStackTrace();
            }
        } finally {
            if ( completed )
                txn.commit();
            else
                txn.abort();
            closeContainer(dbName);
        }
    }

    public static void add(String docString, String docName, String dbName) throws Exception {

        // Need an update context for the put.
        //XmlUpdateContext theContext = XmlManager.createUpdateContext();

        XmlTransaction txn = XMLDB.XMLDBFactory.getXmlManager().createTransaction();
        boolean completed = false;
        try {
            getContainer(dbName).putDocument(txn, docName,     // The document's name
                    docString,   // The actual document, // in a string.
                    null,  // The update context (required).
                    getDocConfig());
            completed = true;
        } finally {
            if ( completed )
                txn.commit();
            else
                txn.abort();
            closeContainer(dbName);
        }
    }

    public static void processResults(XmlResults results, Writer os) throws Exception {

        if ( results == null )
            return;

        XmlValue value = results.next();

        char buffer[] = new char[10000];
        int bytesTotal = 0;

        while (value != null) {
            //System.out.println(value.asString());
            Reader is = new StringReader(value.asString());
            for(;;){
                int numRead = is.read(buffer);
                if ( numRead >0 ){
                    os.write(buffer,0,numRead);
                    bytesTotal += numRead;
                    assertLimit(bytesTotal);
                } else
                    break;
            }
            value = results.next();
        }
    }

    public static String processResults(XmlResults results) throws Exception {
        XmlValue value = results.next();

        StringBuilder results_str = new StringBuilder();
        while (value != null) {
            //System.out.println(value.asString());
            results_str.append("\n");
            results_str.append(value.asString());
            value = results.next();
        }

        return results_str.toString();
    }

    private static void assertLimit(int bytesTotal) throws Exception {
        if (bytesTotal > 20000000 )
            throw new Exception("Output document has exceeded allowed length of 20Mb");
    }

    public static String query(String xq, String dbName) throws Exception {
        String results_str;
        System.err.println(xq);

        XmlQueryContext context = XMLDB.XMLDBFactory.getXmlManager().createQueryContext();
        context.setQueryTimeoutSeconds(100);

        XmlTransaction txn = XMLDB.XMLDBFactory.getXmlManager().createTransaction(null, getTransactionConfig());

        boolean completed = false;
        XmlQueryExpression qe = null;
        XmlResults results = null;
        try {
            openContainer(dbName);
            qe = XMLDB.XMLDBFactory.getXmlManager().prepare(txn, xq, context);
            results = qe.execute(context, getLockedDocConfig());
            results_str = processResults(results);
            completed = true;
        } finally {
            if ( qe != null ) {
                qe.delete();
            }
            if ( results != null ) {
                results.delete();
            }
            if ( completed ) {
                txn.commit();
            }
            else {
                txn.abort();
            }
            closeContainer(dbName);
        }

        return results_str;
    }

    public static XmlContainer openContainer(String dbName) {
        return getContainer(dbName, false);
    }

    public static void closeContainer(String dbName) {
        close(dbName);
    }

    public static synchronized void close(String dbName) {
        //TODO: check if open or not
        try {
            getContainer(dbName).close();
        } catch (XmlException xmle) { //TODO: Propagate exceptions.
            System.err.println("Could not close XmlContainer.");
            xmle.printStackTrace();
        }

    }

    public static XmlContainer createContainer (String dbName) {
        return getContainer(dbName, true);
    }

    public static XmlContainer getContainer(String dbName) {
        return getContainer(dbName, true);
    }

    public static synchronized XmlContainer getContainer(String dbName, boolean create) {

        XmlContainer container = XmlContainerMap.get(dbName);
        
        if(container!=null) {
            return container;
        }

        XmlManager xmlMgr = XMLDB.XMLDBFactory.getXmlManager();

        try {
            if ( xmlMgr.existsContainer(dbName)==0 && create ){
                container = xmlMgr.createContainer(dbName, getContainerConfig());
            } else {
                container = xmlMgr.openContainer(dbName, getContainerConfig());
            }
        } catch (XmlException xmle) { //TODO: Propagate exceptions.
            System.err.println("Could not get XmlContainer.");
            xmle.printStackTrace();
        }

        XmlContainerMap.put(dbName, container);

        return container;
    }

    public static synchronized XmlContainerConfig getContainerConfig() {
        if(containerCfg==null) {
            containerCfg = new XmlContainerConfig();
            containerCfg.setAllowCreate(false);
            containerCfg.setTransactional(true);
        }
        return containerCfg;
    }

    public static synchronized XmlDocumentConfig getDocConfig() {
        if(docConfig==null) {
            docConfig = new XmlDocumentConfig();
            //docConfig.setGenerateName(true);
        }
        return docConfig;
    }

    public static synchronized XmlDocumentConfig getLockedDocConfig() {
        if(docConfig==null) {
            docConfig = new XmlDocumentConfig();
            docConfig.setLockMode(com.sleepycat.db.LockMode.READ_COMMITTED);
        }
        return docConfig;
    }

    public static synchronized TransactionConfig getTransactionConfig() {
        if(txnConfig==null) {
            txnConfig = new TransactionConfig();
            //txnConfig.setSnapshot(true);
            txnConfig.setReadCommitted(true);
        }
        return txnConfig;
    }
}
