xquery version "1.0";
(: Functions for constructing xqueries on XML-formatted Glue 1.3 schema BDII LDAP data. :)

module namespace osg="http://opensciencegrid.org";

import module namespace functx="http://www.functx.com" at "./functx-1.0-nodoc-2007-01.xquery";

(: Returns a string containing a csv list of fqans which match :)
declare function osg:acbr_matchlist($acbr as xs:string, $fqanlist as xs:string) as xs:string {
    let $parts := if (string-length($fqanlist)=0) then "" else osg:acbr_matchlist_untrunc($acbr, $fqanlist)
    let $m0 := replace($parts, ',nomatch', '')
    let $m1 := replace($m0, 'nomatch,', '')
    let $nocomma := if (ends-with($m1, ",")) then substring($m1, 1, string-length($m1)-1) else $m1
    return $nocomma
};

(: Returns a string containing a csv list of fqans which match any node :)
declare function osg:acbr_nodes_matchlist($acbr_nodes as node()*, $fqanlist as xs:string) as xs:string {
    let $list := osg:acbr_nodes_matchlist_untrunc($acbr_nodes, $fqanlist)
    let $nocomma := if (ends-with($list, ",")) then substring($list, 1, string-length($list)-1) else $list
    return $nocomma
};


(: Returns a string containing a csv list of fqans which match any node, and trailing comma :)
declare function osg:acbr_nodes_matchlist_untrunc($acbr_nodes as node()*, $fqanlist as xs:string) as xs:string {
    string-join(
    for $acbr in $acbr_nodes
    return
      osg:acbr_matchlist_untrunc($acbr, $fqanlist)
    ,"")
};

(: Returns a string containing a csv list of fqans which match the node, and trailing comma :)
declare function osg:acbr_matchlist_untrunc($acbr_csv as xs:string, $fqanlist as xs:string) as xs:string {
    if (string-length($fqanlist)=0) then
      string-join(
        let $acbr_split := tokenize($acbr_csv, "&#44;")
        for $acbr in $acbr_split
	  let $lasthalf := functx:substring-after-match(concat("x", $acbr), "x\s*(((VO)|(VOMS)):\s*/?)?")
          let $bareacbr := functx:substring-before-match(concat($lasthalf, "x"), "(/Role=\*)?\s*x")
        return
          concat($bareacbr, "&#44;")
      ,"")
    else
      string-join(
      let $fqan_split := tokenize($fqanlist, "&#44;")
      for $fqan in $fqan_split
      return
        if (osg:acbr_match($acbr_csv, $fqan)) then
          concat($fqan, "&#44;")
        else ()
      ,"")
};


(: Returns true if any VO or role in the comma-separated list matches :)
declare function osg:acbr_match_any($acbr as xs:string, $fqanlist as xs:string) as xs:boolean {
    string-length($fqanlist)=0 or
    matches(osg:acbr_match_any_TF($acbr, $fqanlist), 'T')
};

(: Returns a string containing T or F for each fqan depending on whether it matches :)
declare function osg:acbr_match_any_TF($acbr as xs:string, $fqanlist as xs:string) as xs:string {
    string-join(
    let $fqan_split := tokenize($fqanlist, "&#44;")
    for $fqan in $fqan_split
    return
      if (osg:acbr_match($acbr, $fqan)) then
        "T"
      else
        "F"
    ,"")
};

(: Returns true if VO or role is in comma-separated list, with possible "VO" or "VOMS" prefix :)
declare function osg:acbr_match_dep($acbr as xs:string, $fqan as xs:string) as xs:boolean {
    string-length(osg:acbr_matched_part($acbr, $fqan))>0
};

(: Returns true if VO or role is in ascii #44-separated list, with possible "VO" or "VOMS" prefix :)
declare function osg:acbr_match($acbr as xs:string, $fqan as xs:string) as xs:boolean {
    string-length($fqan)=0 or
    matches($acbr, concat("(^|&#44;)\s*(((VO)|(VOMS)):\s*/?)?", $fqan, "(/Role=\*)?\s*($|&#44;)"), "i")
};

(: Returns the part that matches a VO or role from a comma-separated list, with possible "VO" or "VOMS" prefix :)
declare function osg:acbr_matched_part($acbr as xs:string, $fqan as xs:string) as xs:string {
    let $verbose_fqan := concat("(^|&#44;)\s*(((VO)|(VOMS)):\s*/?)?", $fqan, "(/Role=\*)?\s*($|&#44;)")      
    let $parts := if (string-length($fqan)=0) then $acbr else functx:get-matches($acbr, $verbose_fqan)
    let $firstone := if (count($parts)>1) then subsequence($parts, 2, 1) else $parts
    return $firstone
};

(: Returns true if any VO or role in the comma-separated list is matched in any of the nodes :)
declare function osg:acbr_nodes_match_any($acbr_nodes as node()*, $fqanlist as xs:string) as xs:boolean {
    string-length($fqanlist)=0 or
    matches(osg:acbr_nodes_match_any_TF($acbr_nodes, $fqanlist), 'T')
};

(: Returns a string containing T or F for each node depending on whether it matches :)
declare function osg:acbr_nodes_match_any_TF($acbr_nodes as node()*, $fqanlist as xs:string) as xs:string {
    string-join(
    for $acbr in $acbr_nodes
    return
      if (osg:acbr_match_any($acbr, $fqanlist)) then
        "T"
      else
        "F"
    ,"")
};

(: Returns true if VO or role is matched in any of the nodes :)
declare function osg:acbr_nodes_match($acbr_nodes as node()*, $fqan as xs:string) as xs:boolean {
    string-length($fqan)=0 or
    matches(osg:acbr_nodes_match_TF($acbr_nodes, $fqan), 'T')
};

(: Returns a string containing T or F for each node depending on whether it matches :)
declare function osg:acbr_nodes_match_TF($acbr_nodes as node()*, $fqan as xs:string) as xs:string {
    string-join(
    for $acbr in $acbr_nodes
    return
      if (osg:acbr_match($acbr, $fqan)) then
        "T"
      else
        "F"
    ,"")
};

(: Cleans up possible Role=* or Capability=* or leading slash in input fqan or vo :)
declare function osg:clean_fqan($vo as xs:string) as xs:string {
    let $vo1 := fn:replace($vo, "((/Role=\*)|(/Role=\*/Capability=\*)){1}?\s*$", "", "i")
    return fn:replace($vo1, "^/", "", "i")
};        


(: From chunk key or foreign key, finds first match to id in ascii#44-separated attribute :)
declare function osg:get_attr_1st_match($attr as xs:string, $id as xs:string) as xs:string {
    let $aft_id := substring-after($attr, $id)
    let $aft_sep := substring-after($aft_id, "&#44;")
    let $has_sep := string-length($aft_sep)!=0   
    let $val := if ($has_sep) then substring-before($aft_id, concat("&#44;",$aft_sep)) else $aft_id
    return $val
};

declare function osg:get_attr_csv_match($attr as xs:string, $id as xs:string) as xs:string {
    let $unt := osg:get_attr_csv_match_untrunc($attr, $id)
    return
      if (ends-with($unt, ",")) then substring($unt, 1, string-length($unt)-1) else $unt
};

declare function osg:get_attr_csv_match_untrunc($attr as xs:string, $id as xs:string) as xs:string {
  string-join(
    let $wo_split := tokenize($attr, "&#44;")
    for $val in $wo_split
    return
      if (matches($val, concat(".*?",$id,".*?"))) then
        concat($val, ",")
      else()
  ,"")
};

declare function osg:union_csv_ordered($csv1 as xs:string, $csv2 as xs:string) as xs:string {
    let $unt := osg:union_csv_ordered_untrunc($csv1, $csv2)
    return
      if (ends-with($unt, ",")) then substring($unt, 1, string-length($unt)-1) else $unt
};

declare function osg:union_csv_ordered_untrunc($csv1 as xs:string, $csv2 as xs:string) as xs:string {
  string-join(
  for $val in fn:distinct-values(functx:sort(osg:insert_csv(osg:insert_csv((), $csv1), $csv2)))
    return
      concat($val, ",")
  ,"") 
};

declare function osg:insert_csv($nodes as xs:string*, $cvs as xs:string) as xs:string* {
    let $csv_split := tokenize($cvs, "&#44;")
    for $val in $csv_split
      return
      insert-before($nodes, 1, $val)
};

declare function osg:to_csv($nodes as xs:string*) as xs:string {
    let $unt := osg:to_csv_untrunc($nodes)
    return
      if (ends-with($unt, ",")) then substring($unt, 1, string-length($unt)-1) else $unt
};

declare function osg:to_csv_untrunc($nodes as xs:string*) as xs:string {
    string-join(
    for $val in $nodes
    return
      concat($val, ",")
  ,"")
};
