package gov.fnal.xpathsearch.cache.xmldb;

import com.sleepycat.dbxml.*;
import com.sleepycat.db.Environment;
import com.sleepycat.db.LockDetectMode;
import com.sleepycat.db.EnvironmentConfig;
import com.sleepycat.db.TransactionConfig;

import java.io.File;
import java.io.Writer;
import java.io.Reader;
import java.io.StringReader;
import java.util.Map;


public class XMLDB {

    private String dbName;
    private XmlManager XmlManager;
    private XmlContainer myContainer;
    private XmlDocumentConfig docCfg;

    public static class XMLDBFactory {

        private static XmlManager factoryXmlManager=null;
        private static LockDetector lockDetecor=null;
        private static String dbPath = System.getProperty("XMLDB.home", System.getProperty("user.dir")) + "/db";

        private static XMLDBFactory instance = null;

        static class LockDetector extends Thread {
            private Environment env;
            private boolean terminate;

            LockDetector(Environment env){
                this.env = env;
                this.terminate = false;
                start();
            }
            public void terminate(){ terminate = true; }
            public void run(){
                try {
                    for(;!terminate;){
                        env.detectDeadlocks(LockDetectMode.DEFAULT);
                        Thread.sleep(500);
                    }
                } catch ( Exception ex){

                }
            }
        }
        public static XMLDBFactory getInstance()  throws Exception {
            if ( instance == null ) {
                instance = new XMLDBFactory();
            }
            return instance;
        }

        private XMLDBFactory() throws Exception {
            EnvironmentConfig envConf = new EnvironmentConfig();
            File envFile = new File(dbPath);
            envConf.setAllowCreate(true);         // If the environment does not
            envConf.setInitializeCache(true);     // Turn on the shared memory
            envConf.setInitializeLocking(true);   // Turn on the locking subsystem.
            envConf.setInitializeLogging(true);   // Turn on the logging subsystem.
            envConf.setTransactional(true);       // Turn on the transactional
            envConf.setThreaded(true);
            envConf.setMaxLocks(20000);
            //envConf.setMultiversion(true);

            Environment myEnv = new Environment(envFile, envConf);

            XmlManagerConfig managerConfig = new XmlManagerConfig();
            managerConfig.setAdoptEnvironment(true);
            factoryXmlManager = new XmlManager(myEnv, managerConfig);
            lockDetecor = new LockDetector(myEnv);
        }

        public void close() {
            try {
                instance = null;
                if ( factoryXmlManager != null ) {
                    factoryXmlManager.close();
                }
                if ( lockDetecor != null ) {
                    lockDetecor.terminate();
                }
            } catch ( Exception ex ){}
        }

        public XMLDB create(String partitionName,Object id) throws Exception {
            return new XMLDB(factoryXmlManager, partitionName+id.toString(), true);
        }

        public XMLDB open(String partitionName,Object id) throws Exception {
            return new XMLDB(factoryXmlManager, partitionName+id.toString(), false);
        }

        public XMLDB [] list (String storageClass) { return null; }  //TODO: replace with iterator

        public String query(String xquery) { return null; }

        public static XmlManager getXmlManager() {
            try {
                getInstance();
            } catch (Exception e) {
                System.err.println("Could not get XmlManager instance.");
            }
            return factoryXmlManager;
        }
    }

    protected XMLDB (XmlManager factoryXmlManager, String dbName, boolean create) throws Exception {
        this.XmlManager = factoryXmlManager;

        XmlContainerConfig cfg = new XmlContainerConfig();
        cfg.setAllowCreate(false);
        cfg.setTransactional(true);

        docCfg = new XmlDocumentConfig();
        docCfg.setGenerateName(true);

        if ( create ){
            synchronized ( this ){
                myContainer = XmlManager.createContainer(dbName, cfg);
            }
        } else
            myContainer = XmlManager.openContainer(dbName, cfg);

        this.dbName = dbName;
    }

    public void add(String docString) throws Exception {

        // Need an update context for the put.
        //XmlUpdateContext theContext = XmlManager.createUpdateContext();

        XmlTransaction txn = XmlManager.createTransaction();
        boolean completed = false;
        try {
            myContainer.putDocument(txn,"tmElement",     // The document's name
                    docString,   // The actual document, // in a string.
                    null,  // The update context // (required).
                    docCfg);
            completed = true;
        } finally {
            if ( completed )
                txn.commit();
            else
                txn.abort();
        }
    }

    public void processResults(XmlResults results, Writer os) throws Exception {

        if ( results == null )
            return;

        XmlValue value = results.next();

        char buffer[] = new char[10000];
        int bytesTotal = 0;

        while (value != null) {
            //System.out.println(value.asString());
            Reader is = new StringReader(value.asString());
            for(;;){
                int numRead = is.read(buffer);
                if ( numRead >0 ){
                    os.write(buffer,0,numRead);
                    bytesTotal += numRead;
                    assertLimit(bytesTotal);
                } else
                    break;
            }
            value = results.next();
        }
    }

    private void assertLimit(int bytesTotal) throws Exception {
        if (bytesTotal > 20000000 )
            throw new Exception("Output document has exceeded allowed length of 20Mb");
    }

    public void find(String xq, Writer writer, long from, long to, Map params) throws Exception {

        String myQuery = xq.replaceAll("_doc_","collection('"+dbName+"')");

        if ( to == 0 ) {
            to = System.currentTimeMillis()/1000;
        }

        myQuery = myQuery.replaceAll("_start_",Long.toString(from));
        myQuery = myQuery.replaceAll("_end_",Long.toString(to));

        if ( params != null ){
            for ( Object pName: params.keySet() ){
                String[] paramValues = (String[])params.get(pName);
                myQuery = myQuery.replaceAll("_"+pName+"_",paramValues[0]);
            }
        }

        System.out.println(myQuery);

        XmlQueryContext context = XmlManager.createQueryContext();
        context.setQueryTimeoutSeconds(100);

        XmlDocumentConfig docConfig = new XmlDocumentConfig();
        docConfig.setLockMode(com.sleepycat.db.LockMode.READ_COMMITTED);
        // Perform the query.
        TransactionConfig txnConfig = new TransactionConfig();
        //txnConfig.setSnapshot(true);
        txnConfig.setReadCommitted(true);

        XmlTransaction txn = XmlManager.createTransaction(null,txnConfig);

        boolean completed = false;
        XmlResults results = null;
        try {
            results = XmlManager.query(txn,myQuery, context, docConfig);
            processResults(results,writer);
            completed = true;
        } finally {
            if ( results != null )
                results.delete();
            if ( completed )
                txn.commit();
            else
                txn.abort();
        }
    }

    public void close() {
        try {
            myContainer.close();
        } catch ( Exception ex ){}

    }

    public static void main(String [] args) throws Exception {

        try {
            XMLDB svc = XMLDB.XMLDBFactory.getInstance().create("test1","1234");

            svc.add("<data>DATA</data>");

            svc.close();

        } finally {
            XMLDB.XMLDBFactory.getInstance().close();
        }
    }

}
